from __future__ import annotations
from functools import cached_property
from ..hdlc.frame import Info


class SNRM(Info):
    __max_info_transmit: bytes
    __max_info_receive: bytes
    __window_transmit: bytes
    __window_receive: bytes

    def __init__(self, max_info_transmit: bytes = None,
                 max_info_receive: bytes = None,
                 window_transmit: bytes = None,
                 window_receive: bytes = None):
        self.__max_info_transmit = max_info_transmit
        """ Maximum information field length - transmit """
        self.__max_info_receive = max_info_receive
        """ Maximum information field length - receive """
        self.__window_transmit = window_transmit
        """ Window size k - transmit """
        self.__window_receive = window_receive
        """ Window size k - receive """

    @cached_property
    def content(self) -> bytes:
        value = bytes()
        if self.__max_info_transmit is not None:
            value += b'\x05'
            value += b'\x02'+self.__max_info_transmit if self.__max_info_transmit[0] != 0 \
                else b'\x01' + self.__max_info_transmit[1:]
        if self.__max_info_receive is not None:
            value += b'\x06'  # tag max_value_transmit
            value += b'\x02'+self.__max_info_receive if self.__max_info_receive[0] != 0 \
                else b'\x01' + self.__max_info_receive[1:]
        if self.__window_transmit is not None:
            value += b'\x07\x01' + self.__window_transmit
        if self.__window_receive is not None:
            value += b'\x08\x01'+self.__window_receive
        if len(value) == 0:
            return bytes()
        else:
            return b'\x81\x80' + len(value).to_bytes(1, 'big') + value

    def info(self) -> bytes:
        return self.content

    def __len__(self):
        return len(self.content)

    def __str__(self):
        value: str = ''
        if self.__max_info_transmit:
            value += F'max_tr: {int.from_bytes(self.__max_info_transmit, "big")}'
        if self.__max_info_receive:
            value += F' max_rec: {int.from_bytes(self.__max_info_receive, "big")}'
        if self.__window_transmit:
            value += F' win_tr: {int.from_bytes(self.__window_transmit, "big")}'
        if self.__window_receive:
            value += F' max_rec: {int.from_bytes(self.__window_receive, "big")}'
        return value

    @classmethod
    def try_create(cls, max_info_transmit: bytes = None,
                   max_info_receive: bytes = None,
                   window_transmit: bytes = None,
                   window_receive: bytes = None) -> SNRM | None:
        """ create SNRM if exist as least one parameter """
        if any((max_info_transmit, max_info_receive, window_transmit, window_receive)):
            return cls(max_info_transmit=max_info_transmit,
                       max_info_receive=max_info_receive,
                       window_transmit=window_transmit,
                       window_receive=window_receive)
        else:
            return None



from __future__ import annotations
from struct import pack


class AppVersion:
    __major: int
    __minor: int
    __patch: int
    __additional: str
    __match_args__ = ('major', 'minor', 'patch')

    def __init__(self, major: int, minor: int, patch: int, additional: str = ''):
        self.__major = major
        self.__minor = minor
        self.__patch = patch
        self.__additional = additional

    @classmethod
    def from_str(cls, value: str):
        match value.split(sep='.', maxsplit=2):
            case (major, minor, patch):
                match major.isdigit(), minor.isdigit(), patch.isdigit():
                    case True, True, True:                        return cls(int(major), int(minor), int(patch))
                    case True, True, False:
                        patch_seq = bytearray(patch, 'utf-8')
                        patch_seq.insert(0, ord('0'))  # if not founded integer when patch default value is 0
                        digit_pos = 0
                        while patch_seq[:digit_pos + 1].isdigit() and digit_pos < len(patch_seq):
                            digit_pos += 1
                        return cls(int(major), int(minor), int(patch_seq[:digit_pos]), patch_seq[digit_pos:].decode())
                    case _:                                      return cls(0, 0, 0)
            case _:                                              return cls(0, 0, 0)

    @property
    def major(self) -> int:
        return self.__major

    @property
    def minor(self) -> int:
        return self.__minor

    @property
    def patch(self) -> int:
        return self.__patch

    @property
    def additional(self) -> str:
        return self.__additional

    def __eq__(self, other: AppVersion):
        return hash(self) == hash(other) and self.__additional == other.additional

    def __gt__(self, other: AppVersion):
        return hash(self) > hash(other)

    def __ge__(self, other: AppVersion):
        return hash(self) >= hash(other)

    def __str__(self) -> str:
        return F'{self.__major}.{self.__minor}.{self.__patch}{F" <{self.__additional}>" if len(self.__additional) != 0 else ""}'

    def __hash__(self):
        """ to int6 every 2 byte for major, minor, patch """
        return self.__major * 0x10000_0000 + self.__minor * 0x10000 + self.__patch

    def select_nearest(self, variants: list[AppVersion | str]) -> AppVersion | None:
        """ select left nearest from list. If version is absence return None """
        app_versions: list[AppVersion | None] = [self]
        for version in variants:
            match version:
                case AppVersion(): app_versions.append(version)
                case str():        app_versions.append(self.from_str(version))
                case _:            pass
        app_versions.sort()
        app_versions.append(None)
        return app_versions[app_versions.index(self)-1]


if __name__ == '__main__':
    from time import perf_counter
    a = AppVersion.from_str('1.1.9d1')
    b = AppVersion(3,3,7)
    c = AppVersion(1,3,6)
    d = [a,b,c]
    d1 = sorted(d)
    print(a == b)
    print(b > a)
    match b:
        case AppVersion(3,3, patch) if patch<5:
            print('ok')



from .__class_init__ import *


class CommunicationChannel(cdt.Enum):
    """ ONLY IN CLIENT. Current channel communication. Use with HDLC setup """
    ELEMENTS = {
        b'\x00': en.OPTICAL_PORT,
        b'\x01': en.RS485,
        b'\x02': en.TCP_IP}


class LocalPortNames(cdt.Structure):
    """ Consist names of optical_port and rs-485 by ordering of hdlc_setup """
    values: tuple[cdt.VisibleString, cdt.VisibleString]
    # values: tuple[cdt.Integer, cdt.Integer]
    ELEMENTS = (cdt.StructElement(cdt.se.ITE_OPTICAL_PORT_NAME, cdt.VisibleString),
                cdt.StructElement(cdt.se.ITE_RS485_NAME, cdt.VisibleString))

    @property
    def optical_port_name(self) -> cdt.VisibleString:
        return self.values[0]

    @property
    def rs485_name(self) -> cdt.VisibleString:
        return self.values[1]


class ClientSetup(ic.COSEMInterfaceClasses):
    """ For client setting options """
    NAME = cn.CLIENT_SETUP
    CLASS_ID = ut.CosemClassId(class_id.CLIENT_SETUP)
    VERSION = cdt.Unsigned(0)
    A_ELEMENTS = (ic.ICAElement(an.CHANNEL_COMMUNICATION, CommunicationChannel, default=0),
                  ic.ICAElement(an.SERIAL_PORT_NAMES, LocalPortNames, default=('COM3', 'COM4')),
                  ic.ICAElement(an.CLIENT_NAME, cdt.VisibleString, default="new_device"))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def channel_communication(self) -> CommunicationChannel:
        return self.get_attr(2)

    @property
    def serial_port_names(self) -> LocalPortNames:
        return self.get_attr(3)

    @property
    def client_name(self) -> cdt.VisibleString:
        return self.get_attr(4)


if __name__ == '__main__':
    a = cdt.VisibleString('счетчик №1')
    b = a.decode()
    a = ClientSetup('1.1.1.1.1.1')
    print(a)



import configparser
from . import settings

# set language
config = configparser.ConfigParser()
if config.read("config.ini") and __name__ in config.sections():
    language = config[__name__].get("language")
    settings.set_current_language(language)


from .cosem_interface_classes import collection


CHANNEL_COMMUNICATION = "Канал связи"
SERIAL_PORT_NAMES = "Имена последовательных портов"
CLIENT_NAME = "Имя клиента"

LOGICAL_NAME = "Логическое имя объекта"
VALUE = "Значение"
CALENDAR_NAME_ACTIVE = "Имя активного календаря"
SEASON_PROFILE_ACTIVE = "Таблица активных сезонных профилей"
WEEK_PROFILE_TABLE_ACTIVE = "Таблица активных недельных профилей"
DAY_PROFILE_TABLE_ACTIVE = "Таблица активных суточных профилей"
CALENDAR_NAME_PASSIVE = "Имя пассивного календаря"
SEASON_PROFILE_PASSIVE = "Таблица пассивных сезонных профилей"
ACTIVATE_PASSIVE_CALENDAR_TIME = "Дата активирования пассивного календаря"
WEEK_PROFILE_TABLE_PASSIVE = "Таблица пассивных недельных профилей"
DAY_PROFILE_TABLE_PASSIVE = "Таблица пассивных суточных профилей"
TIME = "Время"
TIME_ZONE = "Часовой пояс"
STATUS = "Статус"
DAYLIGHT_SAVINGS_BEGIN = "Дата перехода на летнее время"
DAYLIGHT_SAVINGS_END = "Дата перехода на зимнее время"
DAYLIGHT_SAVINGS_DEVIATION = "Сдвиг летнего времени"
DAYLIGHT_SAVINGS_ENABLED = "Разрешение перевода на летнее время"
CLOCK_BASE = "Источник времени"
OUTPUT_STATE = "Состояние выключателя"
CONTROL_STATE = "Статус управления"
CONTROL_MODE = "Режим управления"
SCALER_UNIT = "Масштаб и ед. измерения"
CAPTURE_TIME = "Время захвата"
APN = "APN"
PIN_CODE = "ПИН код"
QUALITY_OF_SERVICE = "Качество обслуживания"
OPERATOR = "Оператор"
CS_ATTACHMENT = "Вложения схем переключения"
PS_STATUS = "Статус переключаемых пакетов"
CELL_INFO = "Информация о соединении"
ADJACENT_CELL = "Соседние соединения"
COMM_SPEED = "Скорость обмена"
WINDOWS_SIZE_TRANSMIT = "Размер окна передачи"
WINDOWS_SIZE_RECEIVE = "Размер окна приема"
MAX_INFO_FIELD_LENGTH_TRANSMIT = "Максимальная длина поля данных при передаче"
MAX_INFO_FIELD_LENGTH_RECEIVE = "Максимальная длина поля данных при приеме"
INTER_OCTET_TIME_OUT = "Межсимвольный таймаут"
INACTIVITY_TIME_OUT = "Межкадровый таймаут"
DEVICE_ADDRESS = "Адрес устройства"
IMAGE_BLOCK_SIZE = "Размер блока данных"
IMAGE_TRANSFERRED_BLOCKS_STATUS = "Статус передачи блоков данных"
IMAGE_FIRST_NOT_TRANSFERRED_BLOCK_NUMBER = "Номер первого непереданного блока"
IMAGE_TRANSFER_ENABLED = "Разрешение передачи данных"
IMAGE_TRANSFER_STATUS = "Статус блока данных"
IMAGE_TO_ACTIVATE_INFO = "Информация для активации"
DL_REFERENCE = "Ссылка на данные"
IP_ADDRESS = "IP адрес"
MONITORED_VALUE = "Контролируемая величина"
THRESHOLD_ACTIVE = "Активный порог"
THRESHOLD_NORMAL = "Нормальный порог"
THRESHOLD_EMERGENCY = "Аварийный порог"
MIN_OVER_THRESHOLD_DURATION = "Мин. длительность превышения порога"
MIN_UNDER_THRESHOLD_DURATION = "Мин. длительность снижения ниже порога"
EMERGENCY_PROFILE = "Аварийный профиль"
EMERGENCY_PROFILE_GROUP_ID_LIST = "Список аварийных профилей"
EMERGENCY_PROFILE_ACTIVE = "Активный аварийный профиль"
ACTIONS = "Действия"
BUFFER = "Буфер данных"
CAPTURE_OBJECTS = "Список захватываемых объектов"
CAPTURE_PERIOD = "Период захвата"
SORT_METHOD = "Метод сортировки"
SORT_OBJECT = "Объект сортировки"
ENTRIES_IN_USE = "Занятых записей"
PROFILE_ENTRIES = "Всего записей"
ENTRIES = "Записи"
SCRIPTS = "Сценарий"
EXECUTED_SCRIPT = "Выполненный сценарий"
TYPE = "Тип"
EXECUTION_TIME = "Время исполнения"
TCP_UDP_PORT = "TCP_UDP_порт"
IP_REFERENCE = "IP ссылка"
MMS = "Максимальный размер сегмента"
NB_OF_SIM_CONN = "Кол-во одновременных соединений"
# Настройки безопасности
SECURITY_POLICY = "Политика безопасности"
SECURITY_SUITE = "Комплект безопасности"
CLIENT_SYSTEM_TITLE = "Название клиента"
SERVER_SYSTEM_TITLE = "Название сервера"
CERTIFICATES = "Сертификаты"
# Арбитр
PERMISSION_TABLE = "Таблица разрешений"
WEIGHTINGS_TABLE = "Таблица весов"
MOST_RECENT_REQUESTS_TABLE = "Таблица последних запросов"
LAST_OUTCOME = "Последний результат"
# Регистр контроля
THRESHOLDS = "Пороги"
#
INITIALIZATION_STRING = "Инициализационые строки"
MODEM_PROFILE = "Профиль модема"
OBJECT_LIST = "Список объектов"
ASSOCIATED_PARTNERS_ID = "Идентификатор партнера"
APPLICATION_CONTEXT_NAME = "Имя контекста приложения"
XDLMS_CONTEXT_INFO = "xDLMS информация"
AUTHENTICATION_MECHANISM_NAME = "Имя алгоритма проверки подлинности"
LLS_SECRET = "Секрет (пароль)"
ASSOCIATION_STATUS = "Статус соединения"
SECURITY_SETUP_REFERENCE = "OBIS-код настройки безопасности"
# Push setup
PUSH_OBJECT_LIST = "Список передаваемых объектов"
SEND_DESTINATION_AND_METHOD = "Метод доставки и адресат"
COMMUNICATION_WINDOW = "Окна связи"
RANDOMISATION_START_INTERVAL = "Интервал псевдослучайной задержки выхода"
NUMBER_OF_RETRIES = "Количество повторов"
REPETITION_DELAY = "Задержка повтора"
PORT_REFERENCE = "Настройка порта"
PUSH_CLIENT_SAP = "Номер клиента SAP"
PUSH_PROTECTION_PARAMETERS = "Параметры защиты"
PUSH_OPERATION_METHOD = "Метод работы"
CONFIRMATION_PARAMETERS = "Параметры подтверждения"
LAST_CONFIRMATION_DATE_TIME = "Время последнего подтверждения"



CHANNEL_COMMUNICATION = "channel_communication"
SERIAL_PORT_NAMES = "serial_port_names"
CLIENT_NAME = "client_name"

LOGICAL_NAME = "logical name"
VALUE = "value"
CALENDAR_NAME_ACTIVE = "calendar name active"
SEASON_PROFILE_ACTIVE = "season profile active"
WEEK_PROFILE_TABLE_ACTIVE = "week profile table active"
DAY_PROFILE_TABLE_ACTIVE = "day profile table active"
CALENDAR_NAME_PASSIVE = "calendar name passive"
SEASON_PROFILE_PASSIVE = "season profile passive"
WEEK_PROFILE_TABLE_PASSIVE = "week_profile_table_passive"
DAY_PROFILE_TABLE_PASSIVE = "day_profile_table_passive"
ACTIVATE_PASSIVE_CALENDAR_TIME = "activate passive calendar time"
TIME = "time"
TIME_ZONE = "time_zone"
STATUS = "status"
DAYLIGHT_SAVINGS_BEGIN = "daylight_savings_begin"
DAYLIGHT_SAVINGS_END = "daylight_savings_end"
DAYLIGHT_SAVINGS_DEVIATION = "daylight_savings_deviation"
DAYLIGHT_SAVINGS_ENABLED = "daylight_savings_enabled"
CLOCK_BASE = "clock_base"
OUTPUT_STATE = "output_state"
CONTROL_STATE = "control_state"
CONTROL_MODE = "control_mode"
SCALER_UNIT = "scaler_unit"
CAPTURE_TIME = "capture_time"
APN = "APN"
PIN_CODE = "PIN_code"
QUALITY_OF_SERVICE = "quality_of_service"
OPERATOR = "operator"
CS_ATTACHMENT = "cs_attachment"
PS_STATUS = "ps_status"
CELL_INFO = "cell_info"
ADJACENT_CELL = "adjacent_cell"
COMM_SPEED = "comm_speed"
WINDOWS_SIZE_TRANSMIT = "windows_size_transmit"
WINDOWS_SIZE_RECEIVE = "windows_size_receive"
MAX_INFO_FIELD_LENGTH_TRANSMIT = "max_info_field_length_transmit"
MAX_INFO_FIELD_LENGTH_RECEIVE = "max_info_field_length_receive"
INTER_OCTET_TIME_OUT = "inter_octet_time_out"
INACTIVITY_TIME_OUT = "inactivity_time_out"
DEVICE_ADDRESS = "device_address"
IMAGE_BLOCK_SIZE = "image_block_size"
IMAGE_TRANSFERRED_BLOCKS_STATUS = "image_transferred_blocks_status"
IMAGE_FIRST_NOT_TRANSFERRED_BLOCK_NUMBER = "image_first_not_transferred_block_number"
IMAGE_TRANSFER_ENABLED = "image_transfer_enabled"
IMAGE_TRANSFER_STATUS = "image_transfer_status"
IMAGE_TO_ACTIVATE_INFO = "image_to_activate_info"
DL_REFERENCE = "DL_reference"
IP_ADDRESS = "IP_address"
MONITORED_VALUE = "monitored_value"
THRESHOLD_ACTIVE = "threshold_active"
THRESHOLD_NORMAL = "threshold_normal"
THRESHOLD_EMERGENCY = "threshold_emergency"
MIN_OVER_THRESHOLD_DURATION = "min_over_threshold_duration"
MIN_UNDER_THRESHOLD_DURATION = "min_under_threshold_duration"
EMERGENCY_PROFILE = "emergency_profile"
EMERGENCY_PROFILE_GROUP_ID_LIST = "emergency_profile_group_id_list"
EMERGENCY_PROFILE_ACTIVE = "emergency_profile_active"
ACTIONS = "actions"
BUFFER = "buffer"
CAPTURE_OBJECTS = "capture_objects"
CAPTURE_PERIOD = "capture_period"
SORT_METHOD = "sort_method"
SORT_OBJECT = "sort_object"
ENTRIES_IN_USE = "entries_in_use"
PROFILE_ENTRIES = "profile_entries"
ENTRIES = "entries"
SCRIPTS = "scripts"
EXECUTED_SCRIPT = "executed_script"
TYPE = "type"
EXECUTION_TIME = "execution_time"
TCP_UDP_PORT = "TCP_UDP_port"
IP_REFERENCE = "IP_reference"
MMS = "MMS"
NB_OF_SIM_CONN = "nb_of_sim_conn"
# Security setup
SECURITY_POLICY = "security_policy"
SECURITY_SUITE = "security_suite"
CLIENT_SYSTEM_TITLE = "client_system_title"
SERVER_SYSTEM_TITLE = "server_system_title"
CERTIFICATES = "certificates"
# Arbitrator
PERMISSION_TABLE = "permission_table"
WEIGHTINGS_TABLE = "weightings_table"
MOST_RECENT_REQUESTS_TABLE = "most_recent_requests_table"
LAST_OUTCOME = "last_outcome"
# Register monitor
THRESHOLDS = "thresholds"
#
INITIALIZATION_STRING = "initialization_string"
MODEM_PROFILE = "modem_profile"
OBJECT_LIST = "object_list"
ASSOCIATED_PARTNERS_ID = "associated_partners_id"
APPLICATION_CONTEXT_NAME = "application_context_name"
XDLMS_CONTEXT_INFO = "xDLMS_context_info"
AUTHENTICATION_MECHANISM_NAME = "authentication_mechanism_name"
LLS_SECRET = "LLS_secret"
ASSOCIATION_STATUS = "association_status"
SECURITY_SETUP_REFERENCE = "security_setup_reference"
# Push setup
PUSH_OBJECT_LIST = "push_object_list"
SEND_DESTINATION_AND_METHOD = "send_destination_and_method"
COMMUNICATION_WINDOW = "communication_window"
RANDOMISATION_START_INTERVAL = "randomisation_start_interval"
NUMBER_OF_RETRIES = "number_of_retries"
REPETITION_DELAY = "repetition_delay"
PORT_REFERENCE = "port_reference"
PUSH_CLIENT_SAP = "push_client_SAP"
PUSH_PROTECTION_PARAMETERS = "push_protection_parameters"
PUSH_OPERATION_METHOD = "push_operation_method"
CONFIRMATION_PARAMETERS = "confirmation_parameters"
LAST_CONFIRMATION_DATE_TIME = "last_confirmation_date_time"



from abc import ABC
from itertools import chain
from ..types import common_data_types as cdt, useful_types as ut


class CommonDataTypeChoiceBase(ut.CHOICE, ABC):
    """ For some attributes of some COSEM interface objects, the data type may be chosen at COSEM object instantiation, in the implementation phase
    of the COSEM server. The Server always shall send back the data type and the value of each attribute, so that together with the logical name, an
    unambiguous interpretation is ensured. The list of possible data types is defined in the “Attribute description” section of a COSEM interface
    class specification. DLMS UA 1000-1 Ed. 13. The tag of a type defined using the CHOICE keyword takes the value of the tag of the type from which
    the chosen data value is taken. ITU-T Rec. X.680 | ISO/IEC 8824-1 """
    TYPE = cdt.CommonDataType


class SimpleDataTypeChoice(CommonDataTypeChoiceBase):
    """All Simple Data Type"""
    ELEMENTS = {0: ut.SequenceElement(cdt.tn.NULL_DATA, cdt.NullData),
                3: ut.SequenceElement(cdt.tn.BOOLEAN, cdt.Boolean),
                4: ut.SequenceElement(cdt.tn.BIT_STRING, cdt.BitString),
                5: ut.SequenceElement(cdt.tn.DOUBLE_LONG, cdt.DoubleLong),
                6: ut.SequenceElement(cdt.tn.DOUBLE_LONG_UNSIGNED, cdt.DoubleLongUnsigned),
                9: ut.SequenceElement(cdt.tn.OCTET_STRING, cdt.OctetString),
                10: ut.SequenceElement(cdt.tn.VISIBLE_STRING, cdt.VisibleString),
                12: ut.SequenceElement(cdt.tn.UTF8_STRING, cdt.Utf8String),
                13: ut.SequenceElement(cdt.tn.BCD, cdt.Bcd),
                15: ut.SequenceElement(cdt.tn.INTEGER, cdt.Integer),
                16: ut.SequenceElement(cdt.tn.LONG, cdt.Long),
                17: ut.SequenceElement(cdt.tn.UNSIGNED, cdt.Unsigned),
                18: ut.SequenceElement(cdt.tn.LONG_UNSIGNED, cdt.LongUnsigned),
                20: ut.SequenceElement(cdt.tn.LONG64, cdt.Long64),
                21: ut.SequenceElement(cdt.tn.LONG64_UNSIGNED, cdt.Long64Unsigned),
                22: ut.SequenceElement(cdt.tn.ENUM, cdt.Enum),
                23: ut.SequenceElement(cdt.tn.FLOAT32, cdt.Float32),
                24: ut.SequenceElement(cdt.tn.FLOAT64, cdt.Float64),
                25: ut.SequenceElement(cdt.tn.DATE_TIME, cdt.DateTime),
                26: ut.SequenceElement(cdt.tn.DATE, cdt.Date),
                27: ut.SequenceElement(cdt.tn.TIME, cdt.Time),
                # TODO: and more 28..33
                }


class ComplexDataTypeChoice(CommonDataTypeChoiceBase):
    """All Complex Data Types"""
    ELEMENTS = {0: ut.SequenceElement(cdt.tn.NULL_DATA, cdt.NullData),
                1: ut.SequenceElement(cdt.tn.ARRAY, cdt.Array),
                2: ut.SequenceElement(cdt.tn.STRUCTURE, cdt.Structure),
                19: ut.SequenceElement(cdt.tn.COMPACT_ARRAY, cdt.CompactArray)}


class AccessSelectorsChoice(CommonDataTypeChoiceBase):
    """All Complex Data Types"""
    ELEMENTS = {0: ut.SequenceElement(cdt.tn.NULL_DATA, cdt.NullData),
                1: ut.SequenceElement(cdt.tn.ARRAY, cdt.Array)}


class CommonDataTypeChoice(CommonDataTypeChoiceBase):
    """Types of Data.value"""
    ELEMENTS = dict(chain(SimpleDataTypeChoice.ELEMENTS.items(), ComplexDataTypeChoice.ELEMENTS.items()))


class ExtendedRegisterChoice(CommonDataTypeChoiceBase):
    """Types of ExtendedRegister.value"""
    ELEMENTS = {0: ut.SequenceElement(cdt.tn.NULL_DATA, cdt.NullData),
                4: ut.SequenceElement(cdt.tn.BIT_STRING, cdt.BitString),
                6: ut.SequenceElement(cdt.tn.DOUBLE_LONG_UNSIGNED, cdt.DoubleLongUnsigned),
                9: ut.SequenceElement(cdt.tn.OCTET_STRING, cdt.OctetString),
                10: ut.SequenceElement(cdt.tn.VISIBLE_STRING, cdt.VisibleString),
                12: ut.SequenceElement(cdt.tn.UTF8_STRING, cdt.Utf8String),
                17: ut.SequenceElement(cdt.tn.UNSIGNED, cdt.Unsigned),
                18: ut.SequenceElement(cdt.tn.LONG_UNSIGNED, cdt.LongUnsigned),
                21: ut.SequenceElement(cdt.tn.LONG64_UNSIGNED, cdt.Long64Unsigned)}


class RegisterChoice(CommonDataTypeChoiceBase):
    """Types of ExtendedRegister.value"""
    ELEMENTS = dict(ExtendedRegisterChoice.ELEMENTS)
    ELEMENTS.update({5: ut.SequenceElement(cdt.tn.DOUBLE_LONG, cdt.DoubleLong),
                     15: ut.SequenceElement(cdt.tn.INTEGER, cdt.Integer),
                     16: ut.SequenceElement(cdt.tn.LONG, cdt.Long),
                     20: ut.SequenceElement(cdt.tn.LONG64, cdt.Long64),
                     22: ut.SequenceElement(cdt.tn.ENUM, cdt.Enum),
                     23: ut.SequenceElement(cdt.tn.FLOAT32, cdt.Float32),
                     24: ut.SequenceElement(cdt.tn.FLOAT64, cdt.Float64)})


simple_dt = SimpleDataTypeChoice()
complex_dt = ComplexDataTypeChoice()
common_dt = CommonDataTypeChoice()
extended_register = ExtendedRegisterChoice()
register = RegisterChoice()
access_selectors = AccessSelectorsChoice()



from __future__ import annotations
from ..__class_init__ import *


class SecurityPolicy(cdt.Enum):
    """ Enforces authentication and/or encrypting algorithm provided with security_suite """
    ELEMENTS = {b'\x00': en.NOTHING,
                b'\x01': en.ALL_MESSAGES_TO_BE_AUTHENTICATION,
                b'\x02': en.ALL_MESSAGES_TO_BE_ENCRYPTED,
                b'\x03': en.ALL_MESSAGES_TO_BE_AUTH_AND_ENC,
                b'\x04': en.RESERVED,
                b'\x05': en.RESERVED,
                b'\x06': en.RESERVED,
                b'\x07': en.RESERVED,
                b'\x08': en.RESERVED,
                b'\x09': en.RESERVED,
                b'\x0a': en.RESERVED,
                b'\x0b': en.RESERVED,
                b'\x0c': en.RESERVED,
                b'\x0d': en.RESERVED,
                b'\x0e': en.RESERVED,
                b'\x0f': en.RESERVED}


class SecuritySuite(cdt.Enum):
    """Specifies authentication, encryption and key transport algorithm"""
    ELEMENTS = {b'\x00': en.AES_GCM_128,
                b'\x01': en.RESERVED,
                b'\x02': en.RESERVED,
                b'\x03': en.RESERVED,
                b'\x04': en.RESERVED,
                b'\x05': en.RESERVED,
                b'\x06': en.RESERVED,
                b'\x07': en.RESERVED,
                b'\x08': en.RESERVED,
                b'\x09': en.RESERVED,
                b'\x0a': en.RESERVED,
                b'\x0b': en.RESERVED,
                b'\x0c': en.RESERVED,
                b'\x0d': en.RESERVED,
                b'\x0e': en.RESERVED,
                b'\x0f': en.RESERVED}


class SecurityActivate(cdt.Enum):
    """For Activates and strengthens the security policy"""
    ELEMENTS = {b'\x00': en.NOTHING,
                b'\x01': en.ALL_MESSAGES_TO_BE_AUTHENTICATION,
                b'\x02': en.ALL_MESSAGES_TO_BE_ENCRYPTED,
                b'\x03': en.ALL_MESSAGES_TO_BE_AUTH_AND_ENC}


class KeyID(cdt.Enum):
    """Use only in KeyData structure"""
    ELEMENTS = {b'\x00': en.GUEK,
                b'\x01': en.GBEK,
                b'\x02': en.GAK}


class KeyData(cdt.Structure):
    """ TODO: """
    values: tuple[KeyID, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.KEY_ID, KeyID),
                cdt.StructElement(cdt.se.KEY_WRAPPED, cdt.OctetString))

    @property
    def key_id(self) -> KeyID:
        return self.values[0]

    @property
    def key_wrapped(self) -> cdt.OctetString:
        return self.values[1]


class GlobalKeyTransfer(cdt.Array):
    """ Array of key_data """
    TYPE = KeyData


class SecuritySetup(ic.COSEMInterfaceClasses):
    """ Instances of the “Security setup” IC contain the necessary information on the security suite in use and the security policy applicable between the server and a client
    and/or third party indentify by their respective system titles. They also provide methods to increase the level of security and to manage symmetric keys, asymmetric key pairs
     and certificates """
    NAME = cn.SECURITY_SETUP
    CLASS_ID = ut.CosemClassId(class_id.SECURITY_SETUP)
    VERSION = cdt.Unsigned(0)
    A_ELEMENTS = (ic.ICAElement(an.SECURITY_POLICY, SecurityPolicy, 0, 3, 0),
                  ic.ICAElement(an.SECURITY_SUITE, SecuritySuite, 0, 0, 0),
                  ic.ICAElement(an.CLIENT_SYSTEM_TITLE, cdt.OctetString),
                  ic.ICAElement(an.SERVER_SYSTEM_TITLE, cdt.OctetString))

    M_ELEMENTS = (ic.ICMElement(mn.SECURITY_ACTIVATE, SecurityActivate),
                  ic.ICMElement(mn.GLOBAL_KEY_TRANSFER, GlobalKeyTransfer))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def security_policy(self) -> SecurityPolicy:
        return self.get_attr(2)

    @property
    def security_suite(self) -> SecuritySuite:
        return self.get_attr(3)

    @property
    def client_system_title(self) -> cdt.OctetString:
        return self .get_attr(4)

    @property
    def server_system_title(self) -> cdt.OctetString:
        return self .get_attr(5)

    @property
    def security_activate(self) -> SecurityActivate:
        return self .get_meth(1)

    @property
    def global_key_transfer(self) -> GlobalKeyTransfer:
        return self .get_meth(2)



FAILED_TO_RECEIVE = "Failed to receive reply from the device in given time"
OPEN_PORT = "Open port"
INVALID_PDU = "Invalid PDU"
UPDATE_AVAILABLE = "update available"
UNKNOWN_APPVERSION_TYPE = "Unknown Appversion type"
INITIALIZE_CONNECTION = "Initialize connection"



EXECUTE = "execute"
REPLY_TO_HLS_AUTHENTICATION = "reply_to_HLS_authentication"
CHANGE_HLS_SECRET = "change_HLS_secret"
ADD_OBJECT = "add_object"
REMOVE_OBJECT = "remove_object"
ACTIVATE_PASSIVE_CALENDAR = "activate_passive_calendar"
ADJUST_TO_QUARTER = "adjust_to_quarter"
ADJUST_TO_MEASURING_PERIOD = "adjust_to_measuring_period"
ADJUST_TO_MINUTE = "adjust_to_minute"
ADJUST_TO_PRESET_TIME = "adjust_to_preset_time"
PRESET_ADJUSTING_TIME = "preset_adjusting_time"
SHIFT_TIME = "shift_time"
REMOTE_DISCONNECT = "remote_disconnect"
REMOTE_RECONNECT = "remote_reconnect"
RESET = "reset"
IMAGE_TRANSFER_INITIATE = "image_transfer_initiate"
IMAGE_BLOCK_TRANSFER = "image_block_transfer"
IMAGE_VERIFY = "image_verify"
IMAGE_ACTIVATE = "image_activate"
CAPTURE = "capture"
ENABLE_DISABLE = "enable_disable"
INSERT = "insert"
DELETE = "delete"
SECURITY_ACTIVATE = "security_activate"
GLOBAL_KEY_TRANSFER = "global_key_transfer"
KEY_TRANSFER = "key_transfer"
KEY_AGREEMENT = "key agreement"
GENERATE_KEY_PAIR = "generate key pair"
GENERATE_CERTIFICATE_REQUEST = "generate certificate request"
IMPORT_CERTIFICATE = "import certificate"
EXPORT_CERTIFICATE = "export certificate"
REMOVE_CERTIFICATE = "remove certificate"
REQUEST_ACTION = "request_action"
PUSH = "push"



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Type, Any, Callable
from dataclasses import dataclass
from ..types import common_data_types as cdt


class _String(ABC):
    LENGTH: int | None

    def __init__(self, value: bytes | bytearray | str | int | UsefulType = None):
        match value:
            case None:                                                        self.contents = bytes(self.LENGTH)
            case bytes() if self.LENGTH is None or self.LENGTH <= len(value): self.contents = value[:self.LENGTH]
            # case bytes() if self.LENGTH <= len(value):                       self.contents = value[:self.LENGTH]
            case bytes():            raise ValueError(F'Length of contents for {self.__class__.__name__} must be at least {self.LENGTH}, but got {len(value)}')
            case bytearray():                                                 self.contents = bytes(value)  # Attention!!! changed method content getting from bytearray
            case str():                                                       self.contents = self.from_str(value)
            case int():                                                       self.contents = self.from_int(value)
            case UsefulType():                                                self.contents = value.contents  # TODO: make right type
            case _:                                                           raise ValueError(F'Error create {self.__class__.__name__} with value {value}')

    @abstractmethod
    def __len__(self):
        """ define in subclasses """


class OCTET_STRING(_String):
    """ An ordered sequence of octets (8 bit bytes) """

    def from_str(self, value: str) -> bytes:
        """ input as hex code """
        return bytes.fromhex(value)

    def from_int(self, value: int) -> bytes:
        """ Convert with recursion. Maximum convert length is 32 """
        def to_bytes_with(length_):
            try:
                return int.to_bytes(value, length_, 'big')
            except OverflowError:
                if length_ > 31:
                    raise ValueError(F'Value {value} is big to convert to bytes')
                return to_bytes_with(length_+1)
        length = 1
        return to_bytes_with(length)

    def __str__(self):
        return self.contents.hex(' ')

    def __len__(self):
        return len(self.contents)

    def __getitem__(self, item):
        return self.contents[item]

    # TODO: maybe remove as redundante?
    def decode(self) -> bytes:
        """ decode to build in bytes type """
        return self.contents

    # TODO: maybe remove as redundante?
    def to_str(self, encoding: str = 'cp1251') -> str:
        """ decode to cp1251 by default, replace to '?' if unsupported """
        temp = list()
        for i in self.contents:
            temp.append(i if i > 32 else 63)
        return bytes(temp).decode(encoding)


class CHOICE(ABC):
    """ TODO: with cdt.CHOICE """
    ELEMENTS: dict[int, SequenceElement]

    @property
    @abstractmethod
    def TYPE(self) -> Any:
        """ return valid types """

    def __init_subclass__(cls, **kwargs):
        if hasattr(cls, 'ELEMENTS'):
            for el in cls.ELEMENTS.values():
                if issubclass(el.TYPE, cls.TYPE):
                    """ type in order """
                else:
                    raise ValueError(F'For {cls.__name__} got type {el.TYPE.__name__} with {el.NAME=}, expected {cls.TYPE.__name__}')
        else:
            """ subclass with type carry initiate """

    def __getitem__(self, item: int) -> SequenceElement:
        return self.ELEMENTS[item]

    @property
    def NAME(self) -> str:
        return F'CHOICE[{len(self)}]'

    def __len__(self):
        return len(self.ELEMENTS)

    def is_key(self, value: int) -> bool:
        return value in self.ELEMENTS.keys()

    def __call__(self, value: bytes | int) -> cdt.CommonDataType:
        """ get instance from encoding or tag(with default value). For CommonDataType only """
        try:
            match value:
                case bytes() as encoding: return self.ELEMENTS[encoding[0]].TYPE(encoding)
                case int() as tag:        return self.ELEMENTS[tag].TYPE()
                case _ as error:          raise ValueError(F'Unknown value type {error}')
        except KeyError as e:
            raise KeyError(F'For {self.__class__.__name__} got key: {e.args[0]}, expected {tuple(self.ELEMENTS.keys())}')

    def get_types(self) -> tuple[Type[cdt.CommonDataType]]:
        """ Use in setter attribute.value for validate """
        return tuple((seq_el.TYPE for seq_el in self.ELEMENTS.values()))

    def __str__(self):
        return F'{CHOICE}: {", ".join((el.NAME for el in self.ELEMENTS.values()))}'


def get_instance_and_context(meta: Type[UsefulType], value: bytes) -> tuple[UsefulType, bytes]:
    instance = meta(value)
    return instance, value[len(instance.contents):]


@dataclass(frozen=True)
class SequenceElement:
    NAME: str
    TYPE: Type[UsefulType | SEQUENCE | CHOICE | cdt.CommonDataType]

    def __str__(self):
        return F'{self.NAME}: {self.TYPE.__name__}'


class SEQUENCE(ABC):
    """ TODO: """
    ELEMENTS: tuple[SequenceElement | SEQUENCE, ...]
    values: list[UsefulType]

    def __init__(self, value: bytes | tuple | list | None | SEQUENCE = None):
        self.__dict__['values'] = [None] * len(self.ELEMENTS)
        match value:
            case tuple() | list():
                if len(value) != len(self):
                    raise ValueError(F'Struct {self.__class__.__name__} got length:{len(value)}, expected length:{len(self)}')
                self.from_tuple(value)
            case bytes():           self.from_bytes(value)
            case None:              self.from_default()
            case self.__class__():  self.from_bytes(value.contents)
            case _:                 raise TypeError(F'Value: "{value}" not supported')

    def from_default(self):
        for i in range(len(self)):
            self.values[i] = self.ELEMENTS[i].TYPE()

    def from_bytes(self, value: bytes | bytearray):
        for i, element in enumerate(self.ELEMENTS):
            self.values[i], value = get_instance_and_context(element.TYPE, value)

    def from_tuple(self, value: tuple | list):
        for i, val in enumerate(value):
            self.values[i] = self.ELEMENTS[i].TYPE(val)

    def __str__(self):
        return F'{{{", ".join(map(lambda val: F"{val[0].NAME}: {val[1]}", zip(self.ELEMENTS, self.values)))}}}'

    def __get_index(self, name: str) -> int | None:
        """ get index by name. Return None if not found """
        for i, element in enumerate(self.ELEMENTS):
            if element.NAME == name:
                return i
        else:
            return None

    def __setattr__(self, key, value: UsefulType):
        match self.__get_index(key):
            case int() as i if isinstance(value, self.ELEMENTS[i]):                self.values[i] = value
            case int() as i:                raise ValueError(F'Try assign {key} Type got {value.__class__.__name__}, expected {self.ELEMENTS[i].NAME}')
            case _:                raise ValueError(F'Unsupported change: {key}')

    def __getattr__(self, item: str) -> UsefulType:
        match self.__get_index(item):
            case int() as i: return self.values[i]
            case _:          return self.__dict__[item]

    def __getitem__(self, item: str | int) -> UsefulType:
        """ get element by index or name """
        match item:
            case str() as name:  return self.values[self.__get_index(name)]
            case int() as index: return self.values[index]
            case _:              raise ValueError(F'Unsupported type {item.__class__}')

    def __setitem__(self, key: int, value: UsefulType):
        """ set data to element by index """
        if isinstance(value, self.ELEMENTS[key].TYPE):
            self.values[key] = value
        else:
            raise ValueError(F'Type got {value.__class__.__name__}, expected {self.ELEMENTS[key].TYPE}')

    def __len__(self):
        return len(self.ELEMENTS)

    @property
    def contents(self) -> bytes:
        return b''.join(el.contents for el in self.values)

    @property
    def NAME(self) -> str:
        return F'{self.__class__.__name__}[{len(self)}]'


class UsefulType(ABC):
    """"""
    contents: bytes
    __match_args__ = ('contents',)
    cb_post_set: Callable
    cb_preset: Callable

    @abstractmethod
    def __init__(self, value):
        """ constructor """

    def __eq__(self, other: UsefulType):
        match other:
            case self.__class__(self.contents): return True
            case _:                             return False

    def set_contents_from(self, value: UsefulType | bytes | bytearray | str | int | bool | None):
        new_value = self.__class__(value)
        if hasattr(self, 'cb_preset'):
            self.cb_preset(new_value)
        self.__dict__['contents'] = new_value.contents
        if hasattr(self, 'cb_post_set'):
            self.cb_post_set()

    def __setattr__(self, key, value):
        match key:
            case 'contents' as prop if hasattr(self, 'contents'): raise ValueError(F"Don't support set {prop}")
            case _: super().__setattr__(key, value)


class DigitalMixin(ABC):
    """ Default value is 0 """
    SIGNED: bool
    LENGTH: int
    contents: bytes

    def __init__(self, value: bytes | bytearray | str | int | DigitalMixin = None):
        match value:
            case bytes() if self.LENGTH <= len(value): self.contents = value[:self.LENGTH]
            case bytes():            raise ValueError(F'Length of contents for {self.__class__.__name__} must be at least {self.LENGTH}, but got {len(value)}')
            case bytearray():                          self.contents = bytes(value)  # Attention!!! changed method content getting from bytearray
            case str('-') if self.SIGNED:              self.contents = bytes(self.LENGTH)
            case int():                                self.contents = self.from_int(value)
            case str():                                self.contents = self.from_str(value)
            case None:                                 self.contents = bytes(self.LENGTH)
            case self.__class__():                     self.contents = value.contents
            case _:                                    raise ValueError(F'Error create {self.__class__.__name__} with value: {value}')

    def from_int(self, value: int | float) -> bytes:
        try:
            return int(value).to_bytes(self.LENGTH, 'big', signed=self.SIGNED)
        except OverflowError:
            raise ValueError(F'value {value} out of range')

    def from_str(self, value: str) -> bytes:
        return self.from_int(float(value))

    def __int__(self):
        """ return the build in integer type """
        return int.from_bytes(self.contents, 'big', signed=self.SIGNED)

    def __str__(self):
        return str(int(self))

    def __repr__(self):
        return F'{self.__class__.__name__}({self})'

    def __gt__(self, other: DigitalMixin):
        match other:
            case DigitalMixin(): return int(self) > int(other)
            case _:          raise TypeError(F'Compare type is {other.__class__}, expected Digital')

    def __len__(self) -> int:
        return self.LENGTH

    def __hash__(self):
        return int(self)

    def validate_from(self, value: str, cursor_position: int) -> tuple[str, int]:
        """ return validated value and cursor position. Use in Entry. TODO: remove it, make better """
        type(self)(value=value)
        return value, cursor_position


class Integer8(DigitalMixin, UsefulType):
    """ INTEGER(-127…128) """
    LENGTH = 1
    SIGNED = True


class Integer16(DigitalMixin, UsefulType):
    """ INTEGER(-32 768...32 767) """
    LENGTH = 2
    SIGNED = True


class Integer32(DigitalMixin, UsefulType):
    """ INTEGER(-2 147 483 648...2 147 483 647) """
    LENGTH = 4
    SIGNED = True


class Integer64(DigitalMixin, UsefulType):
    """ INTEGER(-2^63...2^63-1) """
    LENGTH = 8
    SIGNED = True


class Unsigned8(DigitalMixin, UsefulType):
    """ INTEGER(0...255) """
    LENGTH = 1
    SIGNED = False
    # __match_args__ = ('contents',)


class Unsigned16(DigitalMixin, UsefulType):
    """ INTEGER(0...65 535) """
    LENGTH = 2
    SIGNED = False


class Unsigned32(DigitalMixin, UsefulType):
    """ INTEGER(0...4 294 967 295) """
    LENGTH = 4
    SIGNED = False


class Unsigned64(DigitalMixin, UsefulType):
    """ INTEGER(0...264-1) """
    LENGTH = 8
    SIGNED = False


class CosemClassId(Unsigned16):
    """ Identification code of the IC (range 0 to 65 535). The class_id of each object is retrieved together with the logical name by reading the object_list attribute of an
    “Association LN” / ”Association SN” object.
        - class_id-s from 0 to 8 191 are reserved to be specified by the DLMS UA.
        - class_id-s from 8 192 to 32 767 are reserved for manufacturer specific ICs.
        - class_id-s from 32 768 to 65 535 are reserved for user group specific ICs.
    The DLMS UA reserves the right to assign ranges to individual manufacturers or user groups. """


class CosemObjectInstanceId(OCTET_STRING, UsefulType):
    LENGTH = 6

    def __str__(self):
        return '.'.join(map(str, self.contents))

    def from_str(self, value: str) -> bytes:
        """ create logical_name: octet_string from string type ddd.ddd.ddd.ddd.ddd.ddd, ex.: 0.0.1.0.0.255 """
        raw_value = bytes()
        for typecast, separator in zip((self.__from_group_A, )*5+(self.__from_group_F, ), ('.', '.', '.', '.', '.', ' ')):
            try:
                element, value = value.split(separator, 1)
            except ValueError:
                element, value = value, ''
            raw_value += typecast(element)
        return raw_value

    def __from_group_A(self, value: str) -> bytes:
        if isinstance(value, str):
            if value == '':
                return b'\x00'
            try:
                return int(value).to_bytes(1, 'big')
            except OverflowError:
                raise ValueError(F'Int too big to convert {value}')
        else:
            raise TypeError(F'Unsupported type validation from string, got {value.__class__}')

    def __from_group_F(self, value: str) -> bytes:
        if isinstance(value, str):
            if value == '':
                return b'\xff'
            try:
                return int(value).to_bytes(1, 'big')
            except OverflowError:
                raise ValueError(F'Int too big to convert {value}')
        else:
            raise TypeError(F'Unsupported type validation from string, got {value.__class__}')


class CosemObjectAttributeId(Integer8):
    """ TODO """


class CosemObjectMethodId(Integer8):
    """ TODO """


class AccessSelectionParameters(Unsigned8):
    """ Unsigned8(0..1) """

    def __init__(self, value: int | str | Unsigned8 = 1):
        super(AccessSelectionParameters, self).__init__(value)
        if int(self) > 1 or int(self) < 0:
            raise ValueError(F'The {self.__class__.__name__} got {self}, expected 0..1')


class CosemAttributeDescriptor(SEQUENCE):
    class_id: CosemClassId
    instance_id: CosemObjectInstanceId
    attribute_id: CosemObjectAttributeId
    access_selection_parameters: AccessSelectionParameters
    ELEMENTS = (SequenceElement('class_id', CosemClassId),
                SequenceElement('instance_id', CosemObjectInstanceId),
                SequenceElement('attribute_id', CosemObjectAttributeId))

    def __init__(self, value: bytes | tuple | list = None):
        super(CosemAttributeDescriptor, self).__init__(value)
        self.__dict__['access_selection_parameters'] = AccessSelectionParameters(0)

    @property
    def contents(self) -> bytes:
        """ Always contain Access_Selection_Parameters. DLMS UA 1000-2 Ed.9 Excerpt 9.3.9.1.3 """
        return super(CosemAttributeDescriptor, self).contents + self.access_selection_parameters.contents


class CosemMethodDescriptor(SEQUENCE):
    class_id: CosemClassId
    instance_id: CosemObjectInstanceId
    method_id: CosemObjectMethodId
    ELEMENTS = (SequenceElement('class_id', CosemClassId),
                SequenceElement('instance_id', CosemObjectInstanceId),
                SequenceElement('method_id', CosemObjectMethodId))

    def __init__(self, value: tuple[CosemClassId, CosemObjectInstanceId, CosemObjectMethodId]):
        super(CosemMethodDescriptor, self).__init__(value)


class Data(CHOICE, ABC):
    TYPE = cdt.CommonDataType


class SelectiveAccessDescriptor(SEQUENCE, ABC):
    """ Selective access specification always starts with an access selector, followed by an access-specific access parameter list.
    Specified IS/IEC 62056-53 : 2006, 7.4.1.6 Selective access """
    access_selector: Unsigned8
    access_parameters: cdt.CommonDataType
    ELEMENTS: tuple[SequenceElement, SequenceElement]

    def __init__(self, value: tuple | bytes | None = None):
        super(SelectiveAccessDescriptor, self).__init__(value)
        self.access_selector.cb_post_set = self.__validate_selector

    @property
    @abstractmethod
    def ELEMENTS(self) -> tuple[SequenceElement, SequenceElement]:
        """ return elements """

    def from_default(self):
        self.values[0] = self.ELEMENTS[0].TYPE()
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE()

    def from_bytes(self, value: bytes | bytearray):
        self.values[0], value = get_instance_and_context(self.ELEMENTS[0].TYPE, value)
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE(value)

    def from_tuple(self, value: tuple | list):
        self.values[0] = self.ELEMENTS[0].TYPE(value[0])
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE(value[1])

    @property
    def contents(self) -> bytes:
        return self.access_selector.contents+self.access_parameters.encoding

    def __validate_selector(self):
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE()
        print('change sel')

    def set_selector(self, index: int, value=None):
        """ set selector value from int type """
        self.access_selector.set_contents_from(index)
        if value:
            self.access_parameters.set(value)


class CosemAttributeDescriptorWithSelection(SEQUENCE):
    cosem_attribute_descriptor: CosemAttributeDescriptor
    access_selection: SelectiveAccessDescriptor
    ELEMENTS: tuple[CosemAttributeDescriptor, SequenceElement]

    def __init__(self, value: bytes | tuple | list = None):
        super(CosemAttributeDescriptorWithSelection, self).__init__(value)
        self.cosem_attribute_descriptor.access_selection_parameters.set_contents_from(1)

    @property
    @abstractmethod
    def ELEMENTS(self) -> tuple[SequenceElement, SequenceElement]:
        """ return elements. Need initiate in subclass """


class InvokeIdAndPriority(Unsigned8):

    def __init__(self, value: bytes | bytearray | str | int | DigitalMixin = 0b1100_0000):
        super(InvokeIdAndPriority, self).__init__(value)
        if self.contents[0] & 0b00110000:
            raise ValueError(F'For {self.__class__.__name__} set reserved bits')

    @classmethod
    def from_parameters(cls, invoke_id: int = 0,
                        service_class: int = 0,
                        priority: int = 0):
        instance = cls()
        instance.invoke_id = invoke_id
        instance.service_class = service_class
        instance.priority = priority
        return instance

    @property
    def invoke_id(self) -> int:
        return self.contents[0] & 0b0000_1111

    @invoke_id.setter
    def invoke_id(self, value: int):
        if value & 0b00001111:
            self.__dict__['contents'] = int((self.contents[0] & 0b1111_0000) | value).to_bytes(1, 'big')
        else:
            raise ValueError(F'Got {value} for invoke-id, expected 0..15')

    @property
    def service_class(self) -> int:
        """ 0: Unconfirmed or 1: Confirmed bit return """
        return (self.contents[0] >> 6) & 0b1

    @service_class.setter
    def service_class(self, value: int):
        match value:
            case 0 | 1: self.__dict__['contents'] = int((self.contents[0] & 0b1011_1111) | (value << 6)).to_bytes(1, 'big')
            case _:     raise ValueError(F'Got {value} for service_class, expected 0..1')

    @property
    def priority(self) -> int:
        """ 0: Normal or 1: High bit return """
        return (self.contents[0] >> 7) & 0b1

    @priority.setter
    def priority(self, value: int):
        match value:
            case 0 | 1: self.__dict__['contents'] = int((self.contents[0] & 0b0111_1111) | (value << 7)).to_bytes(1, 'big')
            case _:     raise ValueError(F'Got {value} for service_class, expected 0..1')

    def __str__(self):
        return F'priority: {"High" if self.contents[0] & 0b1000_0000 else "Normal"}, ' \
               F'service-class: {"Confirmed" if self.contents[0] & 0b0100_0000 else "Unconfirmed"}, ' \
               F'invoke-id: {self.invoke_id},'


if __name__ == '__main__':
    a = CosemObjectAttributeId(1)
    b = CosemObjectAttributeId(2)
    print(a > b)

    a = CosemClassId(cdt.LongUnsigned(1).contents)
    a = InvokeIdAndPriority()
    c = a.invoke_id
    d = a.service_class
    f = a.priority
    a.service_class = 1
    a.invoke_id = 14
    e = InvokeIdAndPriority.from_parameters(1, 1, 1)
    class AccessSelector(Unsigned8):
        """ Unsigned8 1..4 """
        def __init__(self, value: int | str | Unsigned8 = 1):
            super(AccessSelector, self).__init__(value)
            if int(self) > 4 or int(self) < 1:
                raise ValueError(F'The {self.__class__.__name__} got {self}, expected 1..4')


    class MyData(Data):
        ELEMENTS = {1: SequenceElement('0 a', cdt.NullData),
                    2: SequenceElement('1 d', cdt.Integer),
                    3: SequenceElement('second', cdt.ScalUnitType),
                    4: SequenceElement('3', cdt.Integer)}

    class My(SelectiveAccessDescriptor):
        access_selector: AccessSelector
        access_parameters: MyData
        ELEMENTS = (SequenceElement('access_selector', AccessSelector),
                    SequenceElement('access_parameters', MyData))

    ba = My()
    b = My((3, (10, 10)))
    b2 = b.access_selector
    b3 = b.access_parameters
    b4 = b.access_parameters.unit
    b_repr = My(b'\x03\x0f"')
    b.set_selector(3, 34)
    a = CosemAttributeDescriptor((1, '1.1.1.1.1.1', 1))
    a_repr = CosemAttributeDescriptor(b'\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00')

    class MyWith(CosemAttributeDescriptorWithSelection):
        access_selection: My
        ELEMENTS = (SequenceElement('cosem_attribute_descriptor', CosemAttributeDescriptor),
                    SequenceElement('access_selection', My))

    c = MyWith()
    c_from = MyWith((a, b))
    c_repr = MyWith(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00')
    print(a)
    c = CosemAttributeDescriptor(b'\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00')
    print(c)



from .. import settings
from ..types import common_data_types as cdt


class FlagEvents(dict):
    """ special class for flag events indication """
    def get_report(self, value: int) -> str:
        active_bites = filter(lambda it: value & 2 ** it, range(8))
        names_dict = filter(lambda it_: it_ is not None, map(lambda it: self.get(2 ** it), active_bites))
        language = settings.get_current_language()
        return '. '.join(map(lambda name_dict: name_dict[language], names_dict))


class Events(dict):
    """ special class for enumerate events indication """
    def get_report(self, value: int) -> str:
        match self.get(value):
            case None: return 'error enum'
            case _ as name:    return name[settings.get_current_language()]


# for 0.0.96.11.0.255 События, связанные с напряжением ГОСТ Р 58940—2020 таблица Д.1
voltage_events = Events((
    (1, {settings.Language.ENGLISH: F'Phase L1-voltage drop', settings.Language.RUSSIAN: F'Фаза А - пропадание напряжения'}),
    (2, {settings.Language.ENGLISH: F'Phase L1-voltage recovery', settings.Language.RUSSIAN: F'Фаза А - восстановление напряжения'}),
    (3, {settings.Language.ENGLISH: F'Phase L2-voltage drop', settings.Language.RUSSIAN: F'Фаза В - пропадание напряжения'}),
    (4, {settings.Language.ENGLISH: F'Phase L2-voltage recovery', settings.Language.RUSSIAN: F'Фаза В - восстановление напряжения'}),
    (5, {settings.Language.ENGLISH: F'Phase L3-voltage drop', settings.Language.RUSSIAN: F'Фаза С - пропадание напряжения'}),
    (6, {settings.Language.ENGLISH: F'Phase L3-voltage recovery', settings.Language.RUSSIAN: F'Фаза С - восстановление напряжения'}),
    (7, {settings.Language.ENGLISH: F'Exceeding the voltage of any phase', settings.Language.RUSSIAN: F'Превышение напряжения любой фазы'}),
    (8, {settings.Language.ENGLISH: F'End of over voltage of any phase', settings.Language.RUSSIAN: F'Окончание перенапряжения любой фазы'}),
    (9, {settings.Language.ENGLISH: F'Low voltage of any phase-start', settings.Language.RUSSIAN: F'Низкое напряжение любой фазы - начало'}),
    (10, {settings.Language.ENGLISH: F'Low voltage of any phase-end', settings.Language.RUSSIAN: F'Низкое напряжение любой фазы - окончание'}),
    (11, {settings.Language.ENGLISH: F'Excess of the voltage unbalance coefficient in the reverse sequence-the beginning',
          settings.Language.RUSSIAN: F'Превышение коэффициента несимметрии напряжений по обратной последовательности - начало'}),
    (12, {settings.Language.ENGLISH: F'Excess of the voltage unbalance coefficient in the reverse sequence - end',
          settings.Language.RUSSIAN: F'Превышение коэффициента несимметрии напряжений по обратной последовательности - окончание'}),
    (13, {settings.Language.ENGLISH: F'Phase L1-over voltage start', settings.Language.RUSSIAN: F'Фаза А - перенапряжение начало'}),
    (14, {settings.Language.ENGLISH: F'Phase L1-over voltage end', settings.Language.RUSSIAN: F'Фаза А - перенапряжение окончание'}),
    (15, {settings.Language.ENGLISH: F'Phase L2-over voltage start', settings.Language.RUSSIAN: F'Фаза В - перенапряжение начало'}),
    (16, {settings.Language.ENGLISH: F'Phase L2-over voltage end', settings.Language.RUSSIAN: F'Фаза В - перенапряжение окончание'}),
    (17, {settings.Language.ENGLISH: F'Phase L3-over voltage start', settings.Language.RUSSIAN: F'Фаза С - перенапряжение начало'}),
    (18, {settings.Language.ENGLISH: F'Phase L3-over voltage end', settings.Language.RUSSIAN: F'Фаза С - перенапряжение окончание'}),
    (19, {settings.Language.ENGLISH: F'Phase L1-failure start', settings.Language.RUSSIAN: F'Фаза А - провал начало'}),
    (20, {settings.Language.ENGLISH: F'Phase L1-failure ending', settings.Language.RUSSIAN: F'Фаза А - провал окончание'}),
    (21, {settings.Language.ENGLISH: F'Phase L2-failure start', settings.Language.RUSSIAN: F'Фаза В - провал начало'}),
    (22, {settings.Language.ENGLISH: F'Phase L2-failure ending', settings.Language.RUSSIAN: F'Фаза В - провал окончание'}),
    (23, {settings.Language.ENGLISH: F'Phase L3-failure start', settings.Language.RUSSIAN: F'Фаза С - провал начало'}),
    (24, {settings.Language.ENGLISH: F'Phase L3-failure ending', settings.Language.RUSSIAN: F'Фаза С - провал окончание'}),
    (25, {settings.Language.ENGLISH: F'Incorrect phase sequence start', settings.Language.RUSSIAN: F'Неправильная последовательность фаз начало'}),
    (26, {settings.Language.ENGLISH: F'Incorrect phase sequence ending', settings.Language.RUSSIAN: F'Неправильная последовательность фаз окончание'}),
    (27, {settings.Language.ENGLISH: F'Voltage interruption', settings.Language.RUSSIAN: F'Прерывание напряжения'}),
    (28, {settings.Language.ENGLISH: F'Voltage recovery', settings.Language.RUSSIAN: F'Восстановление напряжения'})))

# for current_events События, связанные с током ГОСТ Р 58940—2020 таблица Д.2
current_events = Events((
    (1, {settings.Language.ENGLISH: F'Phase L1-export start', settings.Language.RUSSIAN: F'Фаза А - экспорт начало'}),
    (2, {settings.Language.ENGLISH: F'Phase L1-Export end', settings.Language.RUSSIAN: F'Фаза А - экспорт окончание'}),
    (3, {settings.Language.ENGLISH: F'Phase L2-export start', settings.Language.RUSSIAN: F'Фаза В - экспорт начало'}),
    (4, {settings.Language.ENGLISH: F'Phase L2-Export end', settings.Language.RUSSIAN: F'Фаза В - экспорт окончание'}),
    (5, {settings.Language.ENGLISH: F'Phase L3-export start', settings.Language.RUSSIAN: F'Фаза С - экспорт начало'}),
    (6, {settings.Language.ENGLISH: F'Phase L3-Export end', settings.Language.RUSSIAN: F'Фаза С - экспорт окончание'}),
    (7, {settings.Language.ENGLISH: F'Phase L1 current transformer breakage', settings.Language.RUSSIAN: F'Обрыв трансформатора тока фазы А'}),
    (8, {settings.Language.ENGLISH: F'Restoring the Phase L1 current transformer', settings.Language.RUSSIAN: F'Восстановление трансформатора тока фазы А'}),
    (9, {settings.Language.ENGLISH: F'Phase L2 current transformer breakage', settings.Language.RUSSIAN: F'Обрыв трансформатора тока фазы В'}),
    (10, {settings.Language.ENGLISH: F'Restoring the Phase L2 current transformer', settings.Language.RUSSIAN: F'Восстановление трансформатора тока фазы В'}),
    (11, {settings.Language.ENGLISH: F'Phase L3 current transformer breakage', settings.Language.RUSSIAN: F'Обрыв трансформатора тока фазы С'}),
    (12, {settings.Language.ENGLISH: F'Restoring the Phase L3 current transformer', settings.Language.RUSSIAN: F'Восстановление трансформатора тока фазы С'}),
    (13, {settings.Language.ENGLISH: F'Current unbalance-the beginning', settings.Language.RUSSIAN: F'Разбаланс токов - начало'}),
    (14, {settings.Language.ENGLISH: F'Current unbalance-end', settings.Language.RUSSIAN: F'Разбаланс токов - окончание'}),
    (15, {settings.Language.ENGLISH: F'Current transformer short circuit-start', settings.Language.RUSSIAN: F'Замыкание трансформатора тока — начало'}),
    (16, {settings.Language.ENGLISH: F'Current transformer short circuit termination', settings.Language.RUSSIAN: F'Окончание замыкания трансформатора тока'}),
    (17, {settings.Language.ENGLISH: F'Exceeding the current of any phase-the beginning', settings.Language.RUSSIAN: F'Превышение тока любой фазы - начало'}),
    (18, {settings.Language.ENGLISH: F'End of excess current of any phase', settings.Language.RUSSIAN: F'Окончание превышения тока любой фазы'}),
    (19, {settings.Language.ENGLISH: F'Phase L1-the presence of current in the absence of voltage start', settings.Language.RUSSIAN: F'Фаза А-наличие тока при отсутствии напряжения начало'}),
    (20, {settings.Language.ENGLISH: F'Phase L1-the presence of current in the absence of voltage end', settings.Language.RUSSIAN: F'Фаза А-наличие тока при отсутствии напряжения окончание'}),
    (21, {settings.Language.ENGLISH: F'Phase L2-the presence of current in the absence of voltage start', settings.Language.RUSSIAN: F'Фаза В - наличие тока при отсутствии напряжения начало'}),
    (22, {settings.Language.ENGLISH: F'Phase L2-the presence of current in the absence of voltage end', settings.Language.RUSSIAN: F'Фаза В - наличие тока при отсутствии напряжения окончание'}),
    (23, {settings.Language.ENGLISH: F'Phase L3-the presence of current in the absence of voltage start', settings.Language.RUSSIAN: F'Фаза С - наличие тока при отсутствии напряжения начало'}),
    (24, {settings.Language.ENGLISH: F'Phase L3-the presence of current in the absence of voltage end', settings.Language.RUSSIAN: F'Фаза С - наличие тока при отсутствии напряжения окончание'}),
    (25, {settings.Language.ENGLISH: F'Phase L1-exceeding the maximum current start', settings.Language.RUSSIAN: F'Фаза А - превышение максимального тока начало'}),
    (26, {settings.Language.ENGLISH: F'Phase L1-exceeding the maximum current end', settings.Language.RUSSIAN: F'Фаза А - превышение максимального тока окончание'}),
    (27, {settings.Language.ENGLISH: F'Phase L2-exceeding the maximum current start', settings.Language.RUSSIAN: F'Фаза В - превышение максимального тока начало'}),
    (28, {settings.Language.ENGLISH: F'Phase L2-exceeding the maximum current end', settings.Language.RUSSIAN: F'Фаза В - превышение максимального тока окончание'}),
    (29, {settings.Language.ENGLISH: F'Phase L3-exceeding the maximum current start', settings.Language.RUSSIAN: F'Фаза С - превышение максимального тока начало'}),
    (30, {settings.Language.ENGLISH: F'Phase L3-exceeding the maximum current end', settings.Language.RUSSIAN: F'Фаза С - превышение максимального тока окончание'}),
    (31, {settings.Language.ENGLISH: F'The presence of current in the absence of voltage (neutral breakage). Begin',
          settings.Language.RUSSIAN: F'Наличие тока при отсутствии напряжения (обрыв нейтрали) – начало'}),
    (32, {settings.Language.ENGLISH: F"The presence of current in the absence of voltage (neutral breakage). End",
          settings.Language.RUSSIAN: F"Наличие тока при отсутствии напряжения (обрыв нейтрали) - окончание"}),
    (33, {settings.Language.ENGLISH: F"Reverse of the power (export of current) in a unidirectional counter. Begin",
          settings.Language.RUSSIAN: F"Обратный поток мощности (экспорт тока) в однонаправленном счётчике - начало"}),
    (34, {settings.Language.ENGLISH: F"Reverse of the power (export of current) in a unidirectional counter. End",
          settings.Language.RUSSIAN: F"Обратный поток мощности (экспорт тока) в однонаправленном счётчике - конец"}),
    (35, {settings.Language.ENGLISH: F"Unspressed phase power in a three -phase and single -phase two -element counter. Begin",
          settings.Language.RUSSIAN: F"Разнонаправленная мощность по фазам в трёхфазном и однофазном двухэлементном счётчике - начало"}),
    (36, {settings.Language.ENGLISH: F"Unspressed phase power in a three -phase and single -phase two -element counter. End",
          settings.Language.RUSSIAN: F"Разнонаправленная мощность по фазам в трёхфазном и однофазном двухэлементном счётчике - окончание"}),
    (37, {settings.Language.ENGLISH: F"The presence of current with the load relay is turned off. Begin",
          settings.Language.RUSSIAN: F"Наличие тока при выключенном реле нагрузки – начало"}),
    (38, {settings.Language.ENGLISH: F"The presence of current with the load relay is turned off. End",
          settings.Language.RUSSIAN: F"Наличие тока при выключенном реле нагрузки – окончание"})))


# for 0.0.96.11.2.255 События, связанные с включением/выключением ПУ. коммутации реле нагрузки ГОСТ Р 58940—2020 таблица Д.3
commutation_events = Events((
    (1, {settings.Language.ENGLISH: F'Power down of the device', settings.Language.RUSSIAN: F'Выключение питания ПУ'}),
    (2, {settings.Language.ENGLISH: F'Powering up the device', settings.Language.RUSSIAN: F'Включение питания ПУ'}),
    (3, {settings.Language.ENGLISH: F'Remote subscriber shutdown', settings.Language.RUSSIAN: F'Выключение абонента дистанционное'}),
    (4, {settings.Language.ENGLISH: F'Enabling the subscriber remotely', settings.Language.RUSSIAN: F'Включение абонента дистанционное'}),
    (5, {settings.Language.ENGLISH: F'Getting permission to enable the subscriber', settings.Language.RUSSIAN: F'Получение разрешения на включение абоненту'}),
    (6, {settings.Language.ENGLISH: F'Switching off the load relay by the subscriber', settings.Language.RUSSIAN: F'Выключение реле нагрузки абонентом'}),
    (7, {settings.Language.ENGLISH: F'Switching on the load relay by the subscriber', settings.Language.RUSSIAN: F'Включение реле нагрузки абонентом'}),
    (8, {settings.Language.ENGLISH: F'Local shutdown when the power limit is exceeded', settings.Language.RUSSIAN: F'Выключение локальное по превышению лимита мощности'}),
    (9, {settings.Language.ENGLISH: F'Local shutdown when the maximum current is exceeded', settings.Language.RUSSIAN: F'Выключение локальное по превышению максимального тока'}),
    (10, {settings.Language.ENGLISH: F'Local shutdown when exposed to a magnetic field', settings.Language.RUSSIAN: F'Выключение локальное при воздействии магнитного поля'}),
    (11, {settings.Language.ENGLISH: F'Local shutdown when the voltage is exceeded', settings.Language.RUSSIAN: F'Выключение локальное по превышению напряжения'}),
    (12, {settings.Language.ENGLISH: F'Switching on local when the voltage returns to normal', settings.Language.RUSSIAN: F'Включение локальное при возвращение напряжения в норму'}),
    (13, {settings.Language.ENGLISH: F'Local shutdown by the presence of current in the absence of voltage',
         settings.Language.RUSSIAN: F'Выключение локальное по наличию тока при отсутствии напряжения'}),
    (14, {settings.Language.ENGLISH: F'Local shutdown based on current unbalance', settings.Language.RUSSIAN: F'Выключение локальное по разбалансу токов'}),
    (15, {settings.Language.ENGLISH: F'Local temperature shutdown', settings.Language.RUSSIAN: F'Выключение локальное по температуре'}),
    (16, {settings.Language.ENGLISH: F'Enabling backup power', settings.Language.RUSSIAN: F'Включение резервного питания'}),
    (17, {settings.Language.ENGLISH: F'Switching off the backup power supply', settings.Language.RUSSIAN: F'Отключение резервного питания'}),
    (18, {settings.Language.ENGLISH: F'Local shutdown when opening the terminal cover or housing',
         settings.Language.RUSSIAN: F'Выключение локальное при вскрытии клеммной крышки или корпуса'})))


# for 0.0.96.11.3.255 События программирования параметров ПУ ГОСТ Р 58940—2020 таблица Д.4
programming_events = Events((
    (1, {settings.Language.ENGLISH: F'Changing the address or the RS-485-1 exchange rate', settings.Language.RUSSIAN: F'Изменение адреса или скорости обмена RS-485-1'}),
    (2, {settings.Language.ENGLISH: F'Changing the address or the RS-485-2 exchange rate', settings.Language.RUSSIAN: F'Изменение адреса или скорости обмена RS-485-2'}),
    (3, {settings.Language.ENGLISH: F'Setting the time', settings.Language.RUSSIAN: F'Установка времени'}),
    (4, {settings.Language.ENGLISH: F'Changing Daylight Saving time settings', settings.Language.RUSSIAN: F'Изменение параметров перехода на летнее время'}),
    (5, {settings.Language.ENGLISH: F'Changing the seasonal profile of the tariff schedule', settings.Language.RUSSIAN: F'Изменение сезонного профиля тарифного расписания (ТР)'}),
    (6, {settings.Language.ENGLISH: F'Changing the weekly profile of the tariff schedule', settings.Language.RUSSIAN: F'Изменение недельного профиля ТР'}),
    (7, {settings.Language.ENGLISH: F'Changing the day profile of the tariff schedule', settings.Language.RUSSIAN: F'Изменение суточного профиля ТР'}),
    (8, {settings.Language.ENGLISH: F'Changing the activation date of the tariff schedule', settings.Language.RUSSIAN: F'Изменение даты активации ТР'}),
    (9, {settings.Language.ENGLISH: F'Activation of the tariff schedule', settings.Language.RUSSIAN: F'Активация ТР'}),
    (10, {settings.Language.ENGLISH: F'Changing the billing day-hour', settings.Language.RUSSIAN: F'Изменение расчетного дня-часа (РДЧ)'}),
    (11, {settings.Language.ENGLISH: F'Changing the display mode (parameters)', settings.Language.RUSSIAN: F'Изменение режима индикации (параметры)'}),
    (12, {settings.Language.ENGLISH: F'Changing the display mode (auto-switch)', settings.Language.RUSSIAN: F'Изменение режима индикации (автопереключение)'}),
    (13, {settings.Language.ENGLISH: F'Changing a low-security (read-only) password)', settings.Language.RUSSIAN: F'Изменение пароля низкой секретности (на чтение)'}),
    (14, {settings.Language.ENGLISH: F'Changing the high-security password (on the record)', settings.Language.RUSSIAN: F'Изменение пароля высокой секретности (на запись)'}),
    (15, {settings.Language.ENGLISH: F'Changing accounting point data', settings.Language.RUSSIAN: F'Изменение данных точки учета'}),
    (16, {settings.Language.ENGLISH: F'Change in the current conversion factor', settings.Language.RUSSIAN: F'Изменение коэффициента трансформации по току'}),
    (17, {settings.Language.ENGLISH: F'Change in the voltage transformation coefficient', settings.Language.RUSSIAN: F'Изменение коэффициента трансформации по напряжению'}),
    (18, {settings.Language.ENGLISH: F'Changing the line parameters for calculating power line losses', settings.Language.RUSSIAN: F'Изменение параметров линии для вычисления потерь в ЛЭП'}),
    (19, {settings.Language.ENGLISH: F'Changing the power limit for disconnection', settings.Language.RUSSIAN: F'Изменение лимита мощности для отключения'}),
    (20, {settings.Language.ENGLISH: F'Changing the power-off time interval', settings.Language.RUSSIAN: F'Изменение интервала времени на отключение по мощности'}),
    (21, {settings.Language.ENGLISH: F'Changing the time interval for switching off when the maximum current is exceeded',
         settings.Language.RUSSIAN: F'Изменение интервала времени на отключение по превышению максимального тока'}),
    (22, {settings.Language.ENGLISH: F'Changing the time interval for switching off at the maximum voltage',
         settings.Language.RUSSIAN: F'Изменение интервала времени на отключение по максимальному напряжению'}),
    (23, {settings.Language.ENGLISH: F'Changing the time interval for switching off due to the influence of a magnetic field',
         settings.Language.RUSSIAN: F'Изменение интервала времени на отключение по воздействию магнитного поля'}),
    (24, {settings.Language.ENGLISH: F'Changing the threshold for fixing a power break', settings.Language.RUSSIAN: F'Изменение порога для фиксации перерыва в питании'}),
    (25, {settings.Language.ENGLISH: F'Changing the threshold for fixing the overvoltage', settings.Language.RUSSIAN: F'Изменение порога для фиксации перенапряжения'}),
    (26, {settings.Language.ENGLISH: F'Changing the threshold for fixing the voltage drop', settings.Language.RUSSIAN: F'Изменение порога для фиксации провала напряжения'}),
    (27, {settings.Language.ENGLISH: F'Changing the threshold for fixing the excess of the tangent', settings.Language.RUSSIAN: F'Изменение порога для фиксации превышения тангенса'}),
    (28, {settings.Language.ENGLISH: F'Changing the threshold for fixing the stress asymmetry coefficient',
         settings.Language.RUSSIAN: F'Изменение порога для фиксации коэффициента несимметрии напряжений'}),
    (29, {settings.Language.ENGLISH: F'Changing the matched voltage', settings.Language.RUSSIAN: F'Изменение согласованного напряжения'}),
    (30, {settings.Language.ENGLISH: F'Changing the peak power integration interval', settings.Language.RUSSIAN: F'Изменение интервала интегрирования пиковой мощности'}),
    (31, {settings.Language.ENGLISH: F'Changing the capture period of profile 1', settings.Language.RUSSIAN: F'Изменение периода захвата профиля 1'}),
    (32, {settings.Language.ENGLISH: F'Changing the capture period of profile 2', settings.Language.RUSSIAN: F'Изменение периода захвата профиля 2'}),
    (33, {settings.Language.ENGLISH: F'Changing the LCD backlight mode', settings.Language.RUSSIAN: F'Изменение режима подсветки LCD'}),
    (34, {settings.Language.ENGLISH: F'Changing the telemetry mode 1', settings.Language.RUSSIAN: F'Изменение режима телеметрии 1'}),
    (35, {settings.Language.ENGLISH: F'Cleaning the monthly log', settings.Language.RUSSIAN: F'Очистка месячного журнала'}),
    (36, {settings.Language.ENGLISH: F'Clearing the daily log', settings.Language.RUSSIAN: F'Очистка суточного журнала'}),
    (37, {settings.Language.ENGLISH: F'Clearing the voltage log', settings.Language.RUSSIAN: F'Очистка журнала напряжения'}),
    (38, {settings.Language.ENGLISH: F'Clearing the current log', settings.Language.RUSSIAN: F'Очистка журнала тока'}),
    (39, {settings.Language.ENGLISH: F'Clearing the on/off log', settings.Language.RUSSIAN: F'Очистка журнала вкл/выкл'}),
    (40, {settings.Language.ENGLISH: F'Clearing the external impact log', settings.Language.RUSSIAN: F'Очистка журнала внешних воздействий'}),
    (41, {settings.Language.ENGLISH: F'Clearing the connection log', settings.Language.RUSSIAN: F'Очистка журнала соединений'}),
    (42, {settings.Language.ENGLISH: F'Clearing the unauthorized access log', settings.Language.RUSSIAN: F'Очистка журнала несанкционированного доступа'}),
    (43, {settings.Language.ENGLISH: F'Clearing the network quality log', settings.Language.RUSSIAN: F'Очистка журнала качества сети'}),
    (44, {settings.Language.ENGLISH: F'Clearing the tangent log', settings.Language.RUSSIAN: F'Очистка журнала тангенса'}),
    (45, {settings.Language.ENGLISH: F'Clearing the I/O log', settings.Language.RUSSIAN: F'Очистка журнала входов/выходов'}),
    (46, {settings.Language.ENGLISH: F'Clearing profile 1', settings.Language.RUSSIAN: F'Очистка профиля 1'}),
    (47, {settings.Language.ENGLISH: F'Clearing profile 2', settings.Language.RUSSIAN: F'Очистка профиля 2'}),
    (48, {settings.Language.ENGLISH: F'Clearing profile 3', settings.Language.RUSSIAN: F'Очистка профиля 3'}),
    (49, {settings.Language.ENGLISH: F'Changing the special Days table', settings.Language.RUSSIAN: F'Изменение таблицы специальных дней'}),
    (50, {settings.Language.ENGLISH: F'Changing the relay control mode', settings.Language.RUSSIAN: F'Изменение режима управления реле'}),
    (51, {settings.Language.ENGLISH: F'Recording of readings in the monthly log', settings.Language.RUSSIAN: F'Фиксация показаний в месячном журнале'}),
    (52, {settings.Language.ENGLISH: F'Changing the initiative exit mode', settings.Language.RUSSIAN: F'Изменение режима инициативного выхода'}),
    (53, {settings.Language.ENGLISH: F'Changing the unicast key for low security', settings.Language.RUSSIAN: F'Изменение одноадресного ключа для низкой секретности'}),
    (54, {settings.Language.ENGLISH: F'Changing the broadcast encryption key for low security',
         settings.Language.RUSSIAN: F'Изменение широковещательного ключа шифрования для низкой секретности'}),
    (55, {settings.Language.ENGLISH: F'Changing the unicast key for high security', settings.Language.RUSSIAN: F'Изменение одноадресного ключа для высокой секретности'}),
    (56, {settings.Language.ENGLISH: F'Changing the broadcast key for high security', settings.Language.RUSSIAN: F'Изменение широковещательного ключа для высокой секретности'}),
    (57, {settings.Language.ENGLISH: F'Changing the authentication key for high security', settings.Language.RUSSIAN: F'Изменение ключа аутентификации для высокой секретности'}),
    (58, {settings.Language.ENGLISH: F'Changing the master key', settings.Language.RUSSIAN: F'Изменение мэстер-ключа'}),
    (59, {settings.Language.ENGLISH: F'Change the conversion level for low secrecy', settings.Language.RUSSIAN: F'Изменение уровня преобразования для низкой секретности'}),
    (60, {settings.Language.ENGLISH: F'Change the conversion level for high security', settings.Language.RUSSIAN: F'Изменение уровня преобразования для высокой секретности'}),
    (61, {settings.Language.ENGLISH: F'Changing the remote display number', settings.Language.RUSSIAN: F'Изменение номера дистанционного дисплея'}),
    (62, {settings.Language.ENGLISH: F'Changing the active energy metering mode (modulo or separately in two directions)',
         settings.Language.RUSSIAN: F'Изменение режима учета активной энергии (по модулю или раздельно в двух направлениях)'}),
    (63, {settings.Language.ENGLISH: F'Setting the time by GPS/GLONASS', settings.Language.RUSSIAN: F'Установка времени по GPS/ГЛОНАСС'}),
    (64, {settings.Language.ENGLISH: F'Changing the neutral break-off mode', settings.Language.RUSSIAN: F'Изменение режима отключения по обрыву нейтрали'}),
    (65, {settings.Language.ENGLISH: F'Software Update', settings.Language.RUSSIAN: F'Обновление ПО'}),
    (66, {settings.Language.ENGLISH: F'Changing the current unbalance shutdown mode', settings.Language.RUSSIAN: F'Изменение режима отключения по разбалансу токов'}),
    (67, {settings.Language.ENGLISH: F'Changing the temperature shutdown mode', settings.Language.RUSSIAN: F'Изменение режима отключения по температуре'}),
    (68, {settings.Language.ENGLISH: F'Time correction', settings.Language.RUSSIAN: F'Коррекция времени'}),
    (69, {settings.Language.ENGLISH: F'Changing the authentication key for low security', settings.Language.RUSSIAN: F'Изменение ключа аутентификации для низкой секретности'}),
    (70, {settings.Language.ENGLISH: F'Clearing Initiative exit flags', settings.Language.RUSSIAN: F'Очистка флагов инициативного выхода'}),
    (71, {settings.Language.ENGLISH: F'Changing the timeout for an HDLC connection', settings.Language.RUSSIAN: F'Изменение таймаута для HDLC-соединения'}),
    (72, {settings.Language.ENGLISH: F'Changing the hours of heavy loads', settings.Language.RUSSIAN: F'Изменение часов болъших нагрузок'}),
    (73, {settings.Language.ENGLISH: F'Changing the maximum monitoring hours', settings.Language.RUSSIAN: F'Изменение часов контроля максимума'}),
    (74, {settings.Language.ENGLISH: F'Changing the connection scheme', settings.Language.RUSSIAN: F'Изменение схемы подключения'}),
    (75, {settings.Language.ENGLISH: F'Changing the telemetry mode 2', settings.Language.RUSSIAN: F'Изменение режима телеметрии 2'}),
    (76, {settings.Language.ENGLISH: F'Changing the telemetry mode 3', settings.Language.RUSSIAN: F'Изменение режима телеметрии 3'}),
    (77, {settings.Language.ENGLISH: F'Changing the telemetry mode 4', settings.Language.RUSSIAN: F'Изменение режима телеметрии 4'}),
    (78, {settings.Language.ENGLISH: F'Changing the shutdown mode when opening the terminal cover or housing',
         settings.Language.RUSSIAN: F'Изменение режима отключения при вскрытии клеммной крышки или корпуса'}),
    (79, {settings.Language.ENGLISH: F'Changing the active communication profile setting for communication ports',
         settings.Language.RUSSIAN: F'Изменение настройки активного коммуникационного профиля для портов связи'}),
    (80, {settings.Language.ENGLISH: F'Clearing the network quality log on a monthly interval', settings.Language.RUSSIAN: F'Очистка журнала качества сети на месячном интервале'}),
    (81, {settings.Language.ENGLISH: F'Changing the interval for integrating network parameters', settings.Language.RUSSIAN: F'Изменение интервала интегрирования параметров сети'}),
    (82, {settings.Language.ENGLISH: F'Changing the threshold value over time. Reactive power factor (tg <p) average for all phases',
         settings.Language.RUSSIAN: F'Изменение порогового значения по времени. Коэффициент реактивной мощности (tg <р) средний по всем фазам'}),
    (83, {settings.Language.ENGLISH: F'Changing the threshold value over time. Differential current. % of the maximum current value',
         settings.Language.RUSSIAN: F'Изменение порогового значения по времени. Дифференциальньм ток. % от величины наибольшего ток'}),
    (84, {settings.Language.ENGLISH: F'Changing the threshold value over time. The coefficient of asymmetry in the reverse sequence',
         settings.Language.RUSSIAN: F'Изменение порогового значения по времени. Коэффициент несимметрии по обратной последовательности'}),
    (85, {settings.Language.ENGLISH: F'Changing the address or exchange rate (Optical port P1)', settings.Language.RUSSIAN: F'Изменение адреса или скорости обмена (Оптопорт Р1)'}),
    (86, {settings.Language.ENGLISH: F'Changing the address or exchange rate (Port P4)', settings.Language.RUSSIAN: F'Изменение адреса или скорости обмена (Порт Р4)'})))


# for 0.0.96.11.4.255 События внешних воздействий ГОСТ Р 58940—2020 таблица Д.5
external_impact_events = Events((
    (1, {settings.Language.ENGLISH: F'Magnetic field - the beginning', settings.Language.RUSSIAN: F'Магнитное поле - начало'}),
    (2, {settings.Language.ENGLISH: F'Magnetic field - ending', settings.Language.RUSSIAN: F'Магнитное поле - окончание'}),
    (3, {settings.Language.ENGLISH: F'Cleansing the electronic seal cover', settings.Language.RUSSIAN: F'Срабатывание электронной пломбы крышки клеммников'}),
    (4, {settings.Language.ENGLISH: F'Plugging of the electoral seal', settings.Language.RUSSIAN: F'Срабатывание электронной пломбы корпуса'})))


# for 0.0.96.11.5.255 Коммуникационные события ГОСТ Р 58940—2020 таблица Д.6
communication_events = Events((
    (1, {settings.Language.ENGLISH: F'Connection terminated (interface)', settings.Language.RUSSIAN: F'Разорвано соединение (интерфейс)'}),
    (2, {settings.Language.ENGLISH: F'Connection established (interface)', settings.Language.RUSSIAN: F'Установлено соединение (интерфейс)'})))

# for 0.0.96.11.6.255 События контроля доступа ГОСТ Р 58940—2020 таблица Д.7
access_events = Events((
    (1, {settings.Language.ENGLISH: F'Unauthorized access attempt (interface)', settings.Language.RUSSIAN: F'Попытка несанкционированного доступа (интерфейс)'}),
    (2, {settings.Language.ENGLISH: F'Violation of the protocol requirements', settings.Language.RUSSIAN: F'Нарушение требований протокола'})))

# for 0.0.96.11.7.255 Коды событий для журнала самодиагностики ГОСТ Р 58940—2020 таблица Д.8
self_diagnostics_events = Events((
    (1, {settings.Language.ENGLISH: F'Initializing the Device', settings.Language.RUSSIAN: F'Инициализация ПУ'}),
    (2, {settings.Language.ENGLISH: F'Measuring unit-error', settings.Language.RUSSIAN: F'Измерительный блок — ошибка'}),
    (3, {settings.Language.ENGLISH: F'Measuring unit-OK', settings.Language.RUSSIAN: F'Измерительный блок — норма'}),
    (4, {settings.Language.ENGLISH: F'Computing block-error', settings.Language.RUSSIAN: F'Вычислительный блок — ошибка'}),
    (5, {settings.Language.ENGLISH: F'Real-time clock-error', settings.Language.RUSSIAN: F'Часы реального времени — ошибка'}),
    (6, {settings.Language.ENGLISH: F'Real-time clock-OK', settings.Language.RUSSIAN: F'Часы реального времени — норма'}),
    (7, {settings.Language.ENGLISH: F'Power supply unit-error', settings.Language.RUSSIAN: F'Блок питания — ошибка'}),
    (8, {settings.Language.ENGLISH: F'Power supply unit-OK', settings.Language.RUSSIAN: F'Блок питания — норма'}),
    (9, {settings.Language.ENGLISH: F'Display-error', settings.Language.RUSSIAN: F'Дисплей — ошибка'}),
    (10, {settings.Language.ENGLISH: F'Display-OK', settings.Language.RUSSIAN: F'Дисплей — норма'}),
    (11, {settings.Language.ENGLISH: F'Memory block-error', settings.Language.RUSSIAN: F'Блок памяти — ошибка'}),
    (12, {settings.Language.ENGLISH: F'Memory block-OK', settings.Language.RUSSIAN: F'Блок памяти — норма'}),
    (13, {settings.Language.ENGLISH: F'Program memory block-error', settings.Language.RUSSIAN: F'Блок памяти программ — ошибка'}),
    (14, {settings.Language.ENGLISH: F'Program memory block-OK', settings.Language.RUSSIAN: F'Блок памяти программ — норма'}),
    (15, {settings.Language.ENGLISH: F'Core clocking system-error', settings.Language.RUSSIAN: F'Система тактирования ядра — ошибка'}),
    (16, {settings.Language.ENGLISH: F'Core clocking system-OK', settings.Language.RUSSIAN: F'Система тактирования ядра — норма'}),
    (17, {settings.Language.ENGLISH: F'Clock clocking system-error', settings.Language.RUSSIAN: F'Система тактирования часов — ошибка'}),
    (18, {settings.Language.ENGLISH: F'Clock clocking system-OK', settings.Language.RUSSIAN: F'Система тактирования часов — норма'}),
    (19, {settings.Language.ENGLISH: F'Computing unit-OK', settings.Language.RUSSIAN: F'Вычислительный блок — норма'}),))

# for 0.0.96.11.8.255 События по превышению реактивной мощности tg (ф) (тангенс сети) ГОСТ Р 58940—2020 таблица Д.9
reactive_power_events = Events((
    (1, {settings.Language.ENGLISH: F'Exceeding the set threshold is the beginning', settings.Language.RUSSIAN: F'Превышение установленного порога — начало'}),
    (2, {settings.Language.ENGLISH: F'Exceeding the set threshold — end', settings.Language.RUSSIAN: F'Превышение установленного порога — окончание'})))

# for 0.0.96.5.1.255 Статус качества сети (профиль суточных показаний)
power_quality_status_1 = FlagEvents((
    (0b00000001, {settings.Language.ENGLISH: F'Deviation of voltage by more than 10% of the nominal', settings.Language.RUSSIAN: F'Отклонение напряжения более чем на 10% от номинала'}),
    (0b00000010, {settings.Language.ENGLISH: F'Frequency deviation by more than 0.4 Hz from the nominal', settings.Language.RUSSIAN: F'Отклонение частоты более чем на 0,4 Гц от номинала'})))

# for 0.0.96.5.4.255 Статус качества сети (журнал качества сети)
power_quality_status_2 = FlagEvents((
    (0b00000001, {settings.Language.ENGLISH: F'Reducing the voltage of more than 10%', settings.Language.RUSSIAN: F'Снижение напряжения более, чем на 10%'}),
    (0b00001000, {settings.Language.ENGLISH: F'Raising voltage for more than 10%', settings.Language.RUSSIAN: F'Повышение напряжения более, чем на 10%'}),
    (0b00010000, {settings.Language.ENGLISH: F'Reducing the frequency of more than 0.4 Hz', settings.Language.RUSSIAN: F'Снижение частоты более, чем на 0,4 Гц'}),
    (0b00100000, {settings.Language.ENGLISH: F'Reducing the frequency of more than 0.2 Hz', settings.Language.RUSSIAN: F'Снижение частоты более, чем на 0,2 Гц'}),
    (0b01000000, {settings.Language.ENGLISH: F'Increase frequency more than 0.2 Hz', settings.Language.RUSSIAN: F'Увеличение частоты более, чем на 0,2 Гц'}),
    (0b10000000, {settings.Language.ENGLISH: F'Increase frequency more than 0.4 Hz', settings.Language.RUSSIAN: F'Увеличение частоты более, чем на 0,4 Гц'})))


def get_report(self, value=None, *args) -> str:
    """ get value of event code by value(attribute 2) in order to event codes table """
    value = self.value if value is None else value
    if isinstance(value, (cdt.Unsigned, cdt.LongUnsigned, cdt.DoubleLongUnsigned, cdt.Long64Unsigned)):
        try:
            return self.events[value.decode()][settings.get_current_language()]
        except KeyError:
            return F'Unknown event code: {value.decode()}'
    else:
        raise TypeError(F'For object {self} wrong value type, got {value.NAME}')



NULL_DATA = "Отсутствие данных"
ARRAY = "Массив"
STRUCTURE = "Структура"
BOOLEAN = "Логические данные"
BIT_STRING = "Последовательность битов"
DOUBLE_LONG = "32-разр целое со знаком"
DOUBLE_LONG_UNSIGNED = "32-разр целое без знака"
OCTET_STRING = "Последовательность байтов"
VISIBLE_STRING = "Последовательность ASCII символов"
UTF8_STRING = "Последовательность символов UTF-8"
BCD = "Двоично-десятичная кодировка байта"
INTEGER = "8-разр целое число со знаком"
LONG = "16-разр целое число со знаком"
UNSIGNED = "8-разр целое число без знака"
LONG_UNSIGNED = "16-разр целое число без знака"
COMPACT_ARRAY = "Массив упакованных данных"
LONG64 = "64-разрядное целое со знаком"
LONG64_UNSIGNED = "64-разрядное целое без знака"
ENUM = "Перечисление"
FLOAT32 = "4-байтовая строка - число с плавающей запятой"
FLOAT64 = "8-байтовая строка - число с плавающей запятой"
DATE_TIME = "Дата-время"
DATE = "Дата"
TIME = "Время"
DON_T_CARE = "dont-care"
